class AbstractCopyTests:
    def test_copy_file_to_existing_directory(
        self, fs, fs_join, fs_bulk_operations_scenario_0, fs_target
    ):
        # Copy scenario 1a
        source = fs_bulk_operations_scenario_0

        target = fs_target
        fs.mkdir(target)
        if not self.supports_empty_directories():
            # Force target directory to exist by adding a dummy file
            fs.touch(fs_join(target, "dummy"))
        assert fs.isdir(target)

        target_file2 = fs_join(target, "file2")
        target_subfile1 = fs_join(target, "subfile1")

        # Copy from source directory
        fs.cp(fs_join(source, "file2"), target)
        assert fs.isfile(target_file2)

        # Copy from sub directory
        fs.cp(fs_join(source, "subdir", "subfile1"), target)
        assert fs.isfile(target_subfile1)

        # Remove copied files
        fs.rm([target_file2, target_subfile1])
        assert not fs.exists(target_file2)
        assert not fs.exists(target_subfile1)

        # Repeat with trailing slash on target
        fs.cp(fs_join(source, "file2"), target + "/")
        assert fs.isdir(target)
        assert fs.isfile(target_file2)

        fs.cp(fs_join(source, "subdir", "subfile1"), target + "/")
        assert fs.isfile(target_subfile1)

    def test_copy_file_to_new_directory(
        self, fs, fs_join, fs_bulk_operations_scenario_0, fs_target
    ):
        # Copy scenario 1b
        source = fs_bulk_operations_scenario_0

        target = fs_target
        fs.mkdir(target)

        fs.cp(
            fs_join(source, "subdir", "subfile1"), fs_join(target, "newdir/")
        )  # Note trailing slash
        assert fs.isdir(target)
        assert fs.isdir(fs_join(target, "newdir"))
        assert fs.isfile(fs_join(target, "newdir", "subfile1"))

    def test_copy_file_to_file_in_existing_directory(
        self, fs, fs_join, fs_bulk_operations_scenario_0, fs_target
    ):
        # Copy scenario 1c
        source = fs_bulk_operations_scenario_0

        target = fs_target
        fs.mkdir(target)

        fs.cp(fs_join(source, "subdir", "subfile1"), fs_join(target, "newfile"))
        assert fs.isfile(fs_join(target, "newfile"))

    def test_copy_file_to_file_in_new_directory(
        self, fs, fs_join, fs_bulk_operations_scenario_0, fs_target
    ):
        # Copy scenario 1d
        source = fs_bulk_operations_scenario_0

        target = fs_target
        fs.mkdir(target)

        fs.cp(
            fs_join(source, "subdir", "subfile1"), fs_join(target, "newdir", "newfile")
        )
        assert fs.isdir(fs_join(target, "newdir"))
        assert fs.isfile(fs_join(target, "newdir", "newfile"))

    def test_copy_directory_to_existing_directory(
        self, fs, fs_join, fs_bulk_operations_scenario_0, fs_target
    ):
        # Copy scenario 1e
        source = fs_bulk_operations_scenario_0

        target = fs_target
        fs.mkdir(target)
        if not self.supports_empty_directories():
            # Force target directory to exist by adding a dummy file
            dummy = fs_join(target, "dummy")
            fs.touch(dummy)
        assert fs.isdir(target)

        for source_slash, target_slash in zip([False, True], [False, True]):
            s = fs_join(source, "subdir")
            if source_slash:
                s += "/"
            t = target + "/" if target_slash else target

            # Without recursive does nothing
            fs.cp(s, t)
            assert fs.ls(target) == [] if self.supports_empty_directories() else [dummy]

            # With recursive
            fs.cp(s, t, recursive=True)
            if source_slash:
                assert fs.isfile(fs_join(target, "subfile1"))
                assert fs.isfile(fs_join(target, "subfile2"))
                assert fs.isdir(fs_join(target, "nesteddir"))
                assert fs.isfile(fs_join(target, "nesteddir", "nestedfile"))
                assert not fs.exists(fs_join(target, "subdir"))

                fs.rm(fs.ls(target, detail=False), recursive=True)
            else:
                assert fs.isdir(fs_join(target, "subdir"))
                assert fs.isfile(fs_join(target, "subdir", "subfile1"))
                assert fs.isfile(fs_join(target, "subdir", "subfile2"))
                assert fs.isdir(fs_join(target, "subdir", "nesteddir"))
                assert fs.isfile(fs_join(target, "subdir", "nesteddir", "nestedfile"))

                fs.rm(fs_join(target, "subdir"), recursive=True)
            assert fs.ls(target) == [] if self.supports_empty_directories() else [dummy]

            # Limit recursive by maxdepth
            fs.cp(s, t, recursive=True, maxdepth=1)
            if source_slash:
                assert fs.isfile(fs_join(target, "subfile1"))
                assert fs.isfile(fs_join(target, "subfile2"))
                assert not fs.exists(fs_join(target, "nesteddir"))
                assert not fs.exists(fs_join(target, "subdir"))

                fs.rm(fs.ls(target, detail=False), recursive=True)
            else:
                assert fs.isdir(fs_join(target, "subdir"))
                assert fs.isfile(fs_join(target, "subdir", "subfile1"))
                assert fs.isfile(fs_join(target, "subdir", "subfile2"))
                assert not fs.exists(fs_join(target, "subdir", "nesteddir"))

                fs.rm(fs_join(target, "subdir"), recursive=True)
            assert fs.ls(target) == [] if self.supports_empty_directories() else [dummy]

    def test_copy_directory_to_new_directory(
        self, fs, fs_join, fs_bulk_operations_scenario_0, fs_target
    ):
        # Copy scenario 1f
        source = fs_bulk_operations_scenario_0

        target = fs_target
        fs.mkdir(target)

        for source_slash, target_slash in zip([False, True], [False, True]):
            s = fs_join(source, "subdir")
            if source_slash:
                s += "/"
            t = fs_join(target, "newdir")
            if target_slash:
                t += "/"

            # Without recursive does nothing
            fs.cp(s, t)
            assert fs.ls(target) == []

            # With recursive
            fs.cp(s, t, recursive=True)
            assert fs.isdir(fs_join(target, "newdir"))
            assert fs.isfile(fs_join(target, "newdir", "subfile1"))
            assert fs.isfile(fs_join(target, "newdir", "subfile2"))
            assert fs.isdir(fs_join(target, "newdir", "nesteddir"))
            assert fs.isfile(fs_join(target, "newdir", "nesteddir", "nestedfile"))
            assert not fs.exists(fs_join(target, "subdir"))

            fs.rm(fs_join(target, "newdir"), recursive=True)
            assert not fs.exists(fs_join(target, "newdir"))

            # Limit recursive by maxdepth
            fs.cp(s, t, recursive=True, maxdepth=1)
            assert fs.isdir(fs_join(target, "newdir"))
            assert fs.isfile(fs_join(target, "newdir", "subfile1"))
            assert fs.isfile(fs_join(target, "newdir", "subfile2"))
            assert not fs.exists(fs_join(target, "newdir", "nesteddir"))
            assert not fs.exists(fs_join(target, "subdir"))

            fs.rm(fs_join(target, "newdir"), recursive=True)
            assert not fs.exists(fs_join(target, "newdir"))

    def test_copy_glob_to_existing_directory(
        self, fs, fs_join, fs_bulk_operations_scenario_0, fs_target
    ):
        # Copy scenario 1g
        source = fs_bulk_operations_scenario_0

        target = fs_target
        fs.mkdir(target)

        for target_slash in [False, True]:
            t = target + "/" if target_slash else target

            # Without recursive
            fs.cp(fs_join(source, "subdir", "*"), t)
            assert fs.isfile(fs_join(target, "subfile1"))
            assert fs.isfile(fs_join(target, "subfile2"))
            assert not fs.isdir(fs_join(target, "nesteddir"))
            assert not fs.exists(fs_join(target, "nesteddir", "nestedfile"))
            assert not fs.exists(fs_join(target, "subdir"))

            fs.rm(fs.ls(target, detail=False), recursive=True)
            assert fs.ls(target) == []

            # With recursive
            fs.cp(fs_join(source, "subdir", "*"), t, recursive=True)
            assert fs.isfile(fs_join(target, "subfile1"))
            assert fs.isfile(fs_join(target, "subfile2"))
            assert fs.isdir(fs_join(target, "nesteddir"))
            assert fs.isfile(fs_join(target, "nesteddir", "nestedfile"))
            assert not fs.exists(fs_join(target, "subdir"))

            fs.rm(fs.ls(target, detail=False), recursive=True)
            assert fs.ls(target) == []

            # Limit recursive by maxdepth
            fs.cp(fs_join(source, "subdir", "*"), t, recursive=True, maxdepth=1)
            assert fs.isfile(fs_join(target, "subfile1"))
            assert fs.isfile(fs_join(target, "subfile2"))
            assert not fs.exists(fs_join(target, "nesteddir"))
            assert not fs.exists(fs_join(target, "subdir"))

            fs.rm(fs.ls(target, detail=False), recursive=True)
            assert fs.ls(target) == []

    def test_copy_glob_to_new_directory(
        self, fs, fs_join, fs_bulk_operations_scenario_0, fs_target
    ):
        # Copy scenario 1h
        source = fs_bulk_operations_scenario_0

        target = fs_target
        fs.mkdir(target)

        for target_slash in [False, True]:
            t = fs_join(target, "newdir")
            if target_slash:
                t += "/"

            # Without recursive
            fs.cp(fs_join(source, "subdir", "*"), t)
            assert fs.isdir(fs_join(target, "newdir"))
            assert fs.isfile(fs_join(target, "newdir", "subfile1"))
            assert fs.isfile(fs_join(target, "newdir", "subfile2"))
            assert not fs.exists(fs_join(target, "newdir", "nesteddir"))
            assert not fs.exists(fs_join(target, "newdir", "nesteddir", "nestedfile"))
            assert not fs.exists(fs_join(target, "subdir"))
            assert not fs.exists(fs_join(target, "newdir", "subdir"))

            fs.rm(fs_join(target, "newdir"), recursive=True)
            assert not fs.exists(fs_join(target, "newdir"))

            # With recursive
            fs.cp(fs_join(source, "subdir", "*"), t, recursive=True)
            assert fs.isdir(fs_join(target, "newdir"))
            assert fs.isfile(fs_join(target, "newdir", "subfile1"))
            assert fs.isfile(fs_join(target, "newdir", "subfile2"))
            assert fs.isdir(fs_join(target, "newdir", "nesteddir"))
            assert fs.isfile(fs_join(target, "newdir", "nesteddir", "nestedfile"))
            assert not fs.exists(fs_join(target, "subdir"))
            assert not fs.exists(fs_join(target, "newdir", "subdir"))

            fs.rm(fs_join(target, "newdir"), recursive=True)
            assert not fs.exists(fs_join(target, "newdir"))

            # Limit recursive by maxdepth
            fs.cp(fs_join(source, "subdir", "*"), t, recursive=True, maxdepth=1)
            assert fs.isdir(fs_join(target, "newdir"))
            assert fs.isfile(fs_join(target, "newdir", "subfile1"))
            assert fs.isfile(fs_join(target, "newdir", "subfile2"))
            assert not fs.exists(fs_join(target, "newdir", "nesteddir"))
            assert not fs.exists(fs_join(target, "subdir"))
            assert not fs.exists(fs_join(target, "newdir", "subdir"))

            fs.rm(fs.ls(target, detail=False), recursive=True)
            assert not fs.exists(fs_join(target, "newdir"))

    def test_copy_list_of_files_to_existing_directory(
        self, fs, fs_join, fs_bulk_operations_scenario_0, fs_target
    ):
        # Copy scenario 2a
        source = fs_bulk_operations_scenario_0

        target = fs_target
        fs.mkdir(target)
        if not self.supports_empty_directories():
            # Force target directory to exist by adding a dummy file
            dummy = fs_join(target, "dummy")
            fs.touch(dummy)
        assert fs.isdir(target)

        source_files = [
            fs_join(source, "file1"),
            fs_join(source, "file2"),
            fs_join(source, "subdir", "subfile1"),
        ]

        for target_slash in [False, True]:
            t = target + "/" if target_slash else target

            fs.cp(source_files, t)
            assert fs.isfile(fs_join(target, "file1"))
            assert fs.isfile(fs_join(target, "file2"))
            assert fs.isfile(fs_join(target, "subfile1"))

            fs.rm(fs.find(target))
            assert fs.ls(target) == [] if self.supports_empty_directories() else [dummy]

    def test_copy_list_of_files_to_new_directory(
        self, fs, fs_join, fs_bulk_operations_scenario_0, fs_target
    ):
        # Copy scenario 2b
        source = fs_bulk_operations_scenario_0

        target = fs_target
        fs.mkdir(target)

        source_files = [
            fs_join(source, "file1"),
            fs_join(source, "file2"),
            fs_join(source, "subdir", "subfile1"),
        ]

        fs.cp(source_files, fs_join(target, "newdir") + "/")  # Note trailing slash
        assert fs.isdir(fs_join(target, "newdir"))
        assert fs.isfile(fs_join(target, "newdir", "file1"))
        assert fs.isfile(fs_join(target, "newdir", "file2"))
        assert fs.isfile(fs_join(target, "newdir", "subfile1"))

    def test_copy_two_files_new_directory(
        self, fs, fs_join, fs_bulk_operations_scenario_0, fs_target
    ):
        # This is a duplicate of test_copy_list_of_files_to_new_directory and
        # can eventually be removed.
        source = fs_bulk_operations_scenario_0

        target = fs_target
        assert not fs.exists(target)
        fs.cp([fs_join(source, "file1"), fs_join(source, "file2")], target)

        assert fs.isdir(target)
        assert fs.isfile(fs_join(target, "file1"))
        assert fs.isfile(fs_join(target, "file2"))



class AbstractGetTests:
    def test_get_file_to_existing_directory(
        self,
        fs,
        fs_join,
        fs_bulk_operations_scenario_0,
        local_fs,
        local_join,
        local_target,
    ):
        # Copy scenario 1a
        source = fs_bulk_operations_scenario_0

        target = local_target
        local_fs.mkdir(target)
        assert local_fs.isdir(target)

        target_file2 = local_join(target, "file2")
        target_subfile1 = local_join(target, "subfile1")

        # Copy from source directory
        fs.get(fs_join(source, "file2"), target)
        assert local_fs.isfile(target_file2)

        # Copy from sub directory
        fs.get(fs_join(source, "subdir", "subfile1"), target)
        assert local_fs.isfile(target_subfile1)

        # Remove copied files
        local_fs.rm([target_file2, target_subfile1])
        assert not local_fs.exists(target_file2)
        assert not local_fs.exists(target_subfile1)

        # Repeat with trailing slash on target
        fs.get(fs_join(source, "file2"), target + "/")
        assert local_fs.isdir(target)
        assert local_fs.isfile(target_file2)

        fs.get(fs_join(source, "subdir", "subfile1"), target + "/")
        assert local_fs.isfile(target_subfile1)

    def test_get_file_to_new_directory(
        self,
        fs,
        fs_join,
        fs_bulk_operations_scenario_0,
        local_fs,
        local_join,
        local_target,
    ):
        # Copy scenario 1b
        source = fs_bulk_operations_scenario_0

        target = local_target
        local_fs.mkdir(target)

        fs.get(
            fs_join(source, "subdir", "subfile1"), local_join(target, "newdir/")
        )  # Note trailing slash

        assert local_fs.isdir(target)
        assert local_fs.isdir(local_join(target, "newdir"))
        assert local_fs.isfile(local_join(target, "newdir", "subfile1"))

    def test_get_file_to_file_in_existing_directory(
        self,
        fs,
        fs_join,
        fs_path,
        fs_bulk_operations_scenario_0,
        local_fs,
        local_join,
        local_target,
    ):
        # Copy scenario 1c
        source = fs_bulk_operations_scenario_0

        target = local_target
        local_fs.mkdir(target)

        fs.get(fs_join(source, "subdir", "subfile1"), local_join(target, "newfile"))
        assert local_fs.isfile(local_join(target, "newfile"))

    def test_get_file_to_file_in_new_directory(
        self,
        fs,
        fs_join,
        fs_bulk_operations_scenario_0,
        local_fs,
        local_join,
        local_target,
    ):
        # Copy scenario 1d
        source = fs_bulk_operations_scenario_0

        target = local_target
        local_fs.mkdir(target)

        fs.get(
            fs_join(source, "subdir", "subfile1"),
            local_join(target, "newdir", "newfile"),
        )
        assert local_fs.isdir(local_join(target, "newdir"))
        assert local_fs.isfile(local_join(target, "newdir", "newfile"))

    def test_get_directory_to_existing_directory(
        self,
        fs,
        fs_join,
        fs_bulk_operations_scenario_0,
        local_fs,
        local_join,
        local_target,
    ):
        # Copy scenario 1e
        source = fs_bulk_operations_scenario_0

        target = local_target
        local_fs.mkdir(target)

        for source_slash, target_slash in zip([False, True], [False, True]):
            s = fs_join(source, "subdir")
            if source_slash:
                s += "/"
            t = target + "/" if target_slash else target

            # Without recursive does nothing
            # ERROR: erroneously creates new directory
            # fs.get(s, t)
            # assert fs.ls(target) == []

            # With recursive
            fs.get(s, t, recursive=True)
            if source_slash:
                assert local_fs.isfile(local_join(target, "subfile1"))
                assert local_fs.isfile(local_join(target, "subfile2"))
                assert local_fs.isdir(local_join(target, "nesteddir"))
                assert local_fs.isfile(local_join(target, "nesteddir", "nestedfile"))

                local_fs.rm(
                    [
                        local_join(target, "subfile1"),
                        local_join(target, "subfile2"),
                        local_join(target, "nesteddir"),
                    ],
                    recursive=True,
                )
            else:
                assert local_fs.isdir(local_join(target, "subdir"))
                assert local_fs.isfile(local_join(target, "subdir", "subfile1"))
                assert local_fs.isfile(local_join(target, "subdir", "subfile2"))
                assert local_fs.isdir(local_join(target, "subdir", "nesteddir"))
                assert local_fs.isfile(
                    local_join(target, "subdir", "nesteddir", "nestedfile")
                )

                local_fs.rm(local_join(target, "subdir"), recursive=True)
            assert local_fs.ls(target) == []

            # Limit by maxdepth
            # ERROR: maxdepth ignored here

    def test_get_directory_to_new_directory(
        self,
        fs,
        fs_join,
        fs_bulk_operations_scenario_0,
        local_fs,
        local_join,
        local_target,
    ):
        # Copy scenario 1f
        source = fs_bulk_operations_scenario_0

        target = local_target
        local_fs.mkdir(target)

        for source_slash, target_slash in zip([False, True], [False, True]):
            s = fs_join(source, "subdir")
            if source_slash:
                s += "/"
            t = local_join(target, "newdir")
            if target_slash:
                t += "/"

            # Without recursive does nothing
            # ERROR: erroneously creates new directory
            # fs.get(s, t)
            # assert fs.ls(target) == []

            # With recursive
            fs.get(s, t, recursive=True)
            assert local_fs.isdir(local_join(target, "newdir"))
            assert local_fs.isfile(local_join(target, "newdir", "subfile1"))
            assert local_fs.isfile(local_join(target, "newdir", "subfile2"))
            assert local_fs.isdir(local_join(target, "newdir", "nesteddir"))
            assert local_fs.isfile(
                local_join(target, "newdir", "nesteddir", "nestedfile")
            )

            local_fs.rm(local_join(target, "newdir"), recursive=True)
            assert local_fs.ls(target) == []

            # Limit by maxdepth
            # ERROR: maxdepth ignored here

    def test_get_glob_to_existing_directory(
        self,
        fs,
        fs_join,
        fs_bulk_operations_scenario_0,
        local_fs,
        local_join,
        local_target,
    ):
        # Copy scenario 1g
        source = fs_bulk_operations_scenario_0

        target = local_target
        local_fs.mkdir(target)

        # for target_slash in [False, True]:
        for target_slash in [False]:
            t = target + "/" if target_slash else target

            # Without recursive
            fs.get(fs_join(source, "subdir", "*"), t)
            assert local_fs.isfile(local_join(target, "subfile1"))
            assert local_fs.isfile(local_join(target, "subfile2"))
            # assert not local_fs.isdir(local_join(target, "nesteddir"))  # ERROR
            assert not local_fs.isdir(local_join(target, "subdir"))

            # With recursive

            # Limit by maxdepth

    def test_get_glob_to_new_directory(
        self,
        fs,
        fs_join,
        fs_bulk_operations_scenario_0,
        local_fs,
        local_join,
        local_target,
    ):
        # Copy scenario 1h
        source = fs_bulk_operations_scenario_0

        target = local_target
        local_fs.mkdir(target)

        for target_slash in [False, True]:
            t = fs_join(target, "newdir")
            if target_slash:
                t += "/"

            # Without recursive
            fs.get(fs_join(source, "subdir", "*"), t)
            assert local_fs.isdir(local_join(target, "newdir"))
            assert local_fs.isfile(local_join(target, "newdir", "subfile1"))
            assert local_fs.isfile(local_join(target, "newdir", "subfile2"))
            # ERROR - do not copy empty directory
            # assert not local_fs.exists(local_join(target, "newdir", "nesteddir"))

            local_fs.rm(local_join(target, "newdir"), recursive=True)
            assert local_fs.ls(target) == []

            # With recursive
            fs.get(fs_join(source, "subdir", "*"), t, recursive=True)
            assert local_fs.isdir(local_join(target, "newdir"))
            assert local_fs.isfile(local_join(target, "newdir", "subfile1"))
            assert local_fs.isfile(local_join(target, "newdir", "subfile2"))
            assert local_fs.isdir(local_join(target, "newdir", "nesteddir"))
            assert local_fs.isfile(
                local_join(target, "newdir", "nesteddir", "nestedfile")
            )

            local_fs.rm(local_join(target, "newdir"), recursive=True)
            assert local_fs.ls(target) == []

            # Limit by maxdepth
            # ERROR: this is not correct

    def test_get_list_of_files_to_existing_directory(
        self,
        fs,
        fs_join,
        fs_bulk_operations_scenario_0,
        local_fs,
        local_join,
        local_target,
    ):
        # Copy scenario 2a
        source = fs_bulk_operations_scenario_0

        target = local_target
        local_fs.mkdir(target)

        source_files = [
            fs_join(source, "file1"),
            fs_join(source, "file2"),
            fs_join(source, "subdir", "subfile1"),
        ]

        for target_slash in [False, True]:
            t = target + "/" if target_slash else target

            fs.get(source_files, t)
            assert local_fs.isfile(local_join(target, "file1"))
            assert local_fs.isfile(local_join(target, "file2"))
            assert local_fs.isfile(local_join(target, "subfile1"))

            local_fs.rm(local_fs.find(target))
            assert local_fs.ls(target) == []

    def test_get_list_of_files_to_new_directory(
        self,
        fs,
        fs_join,
        fs_bulk_operations_scenario_0,
        local_fs,
        local_join,
        local_target,
    ):
        # Copy scenario 2b
        source = fs_bulk_operations_scenario_0

        target = local_target
        local_fs.mkdir(target)

        source_files = [
            fs_join(source, "file1"),
            fs_join(source, "file2"),
            fs_join(source, "subdir", "subfile1"),
        ]

        fs.get(source_files, local_join(target, "newdir") + "/")  # Note trailing slash
        assert local_fs.isdir(local_join(target, "newdir"))
        assert local_fs.isfile(local_join(target, "newdir", "file1"))
        assert local_fs.isfile(local_join(target, "newdir", "file2"))
        assert local_fs.isfile(local_join(target, "newdir", "subfile1"))

    def test_get_directory_recursive(
        self, fs, fs_join, fs_path, local_fs, local_join, local_target
    ):
        # https://github.com/fsspec/filesystem_spec/issues/1062
        # Recursive cp/get/put of source directory into non-existent target directory.
        src = fs_join(fs_path, "src")
        src_file = fs_join(src, "file")
        fs.mkdir(src)
        fs.touch(src_file)

        target = local_target

        # get without slash
        assert not local_fs.exists(target)
        for loop in range(2):
            fs.get(src, target, recursive=True)
            assert local_fs.isdir(target)

        if loop == 0:
            assert local_fs.isfile(local_join(target, "file"))
            assert not local_fs.exists(local_join(target, "src"))
        else:
            assert local_fs.isfile(local_join(target, "file"))
            assert local_fs.isdir(local_join(target, "src"))
            assert local_fs.isfile(local_join(target, "src", "file"))

        local_fs.rm(target, recursive=True)

        # get with slash
        assert not local_fs.exists(target)
        for loop in range(2):
            fs.get(src + "/", target, recursive=True)
            assert local_fs.isdir(target)
            assert local_fs.isfile(local_join(target, "file"))
            assert not local_fs.exists(local_join(target, "src"))



import os

import pytest

from fsspec.implementations.local import LocalFileSystem
from fsspec.tests.abstract.copy import AbstractCopyTests  # noqa
from fsspec.tests.abstract.get import AbstractGetTests  # noqa
from fsspec.tests.abstract.put import AbstractPutTests  # noqa


class BaseAbstractFixtures:
    """
    Abstract base class containing fixtures that are used by but never need to
    be overridden in derived filesystem-specific classes to run the abstract
    tests on such filesystems.
    """

    @pytest.fixture
    def fs_bulk_operations_scenario_0(self, fs, fs_join, fs_path):
        """
        Scenario on remote filesystem that is used for many cp/get/put tests.

        Cleans up at the end of each test it which it is used.
        """
        source = self._bulk_operations_scenario_0(fs, fs_join, fs_path)
        yield source
        fs.rm(source, recursive=True)

    @pytest.fixture
    def fs_target(self, fs, fs_join, fs_path):
        """
        Return name of remote directory that does not yet exist to copy into.

        Cleans up at the end of each test it which it is used.
        """
        target = fs_join(fs_path, "target")
        yield target
        if fs.exists(target):
            fs.rm(target, recursive=True)

    @pytest.fixture
    def local_bulk_operations_scenario_0(self, local_fs, local_join, local_path):
        """
        Scenario on local filesystem that is used for many cp/get/put tests.

        Cleans up at the end of each test it which it is used.
        """
        source = self._bulk_operations_scenario_0(local_fs, local_join, local_path)
        yield source
        local_fs.rm(source, recursive=True)

    @pytest.fixture
    def local_target(self, local_fs, local_join, local_path):
        """
        Return name of local directory that does not yet exist to copy into.

        Cleans up at the end of each test it which it is used.
        """
        target = local_join(local_path, "target")
        yield target
        if local_fs.exists(target):
            local_fs.rm(target, recursive=True)

    def _bulk_operations_scenario_0(self, some_fs, some_join, some_path):
        """
        Scenario that is used for many cp/get/put tests. Creates the following
        directory and file structure:

        📁 source
        ├── 📄 file1
        ├── 📄 file2
        └── 📁 subdir
            ├── 📄 subfile1
            ├── 📄 subfile2
            └── 📁 nesteddir
                └── 📄 nestedfile
        """
        source = some_join(some_path, "source")
        subdir = some_join(source, "subdir")
        nesteddir = some_join(subdir, "nesteddir")
        some_fs.makedirs(nesteddir)
        some_fs.touch(some_join(source, "file1"))
        some_fs.touch(some_join(source, "file2"))
        some_fs.touch(some_join(subdir, "subfile1"))
        some_fs.touch(some_join(subdir, "subfile2"))
        some_fs.touch(some_join(nesteddir, "nestedfile"))
        return source


class AbstractFixtures(BaseAbstractFixtures):
    """
    Abstract base class containing fixtures that may be overridden in derived
    filesystem-specific classes to run the abstract tests on such filesystems.

    For any particular filesystem some of these fixtures must be overridden,
    such as ``fs`` and ``fs_path``, and others may be overridden if the
    default functions here are not appropriate, such as ``fs_join``.
    """

    @pytest.fixture
    def fs(self):
        raise NotImplementedError("This function must be overridden in derived classes")

    @pytest.fixture
    def fs_join(self):
        """
        Return a function that joins its arguments together into a path.

        Most fsspec implementations join paths in a platform-dependent way,
        but some will override this to always use a forward slash.
        """
        return os.path.join

    @pytest.fixture
    def fs_path(self):
        raise NotImplementedError("This function must be overridden in derived classes")

    @pytest.fixture(scope="class")
    def local_fs(self):
        # Maybe need an option for auto_mkdir=False?  This is only relevant
        # for certain implementations.
        return LocalFileSystem(auto_mkdir=True)

    @pytest.fixture
    def local_join(self):
        """
        Return a function that joins its arguments together into a path, on
        the local filesystem.
        """
        return os.path.join

    @pytest.fixture
    def local_path(self, tmpdir):
        return tmpdir

    def supports_empty_directories(self):
        """
        Return whether this implementation supports empty directories.
        """
        return True
